@@ -43,7 +43,7 @@ static int transfer_dir_internal(
   struct vsf_session* p_sess, const int remote_fd,
   struct vsf_sysutil_dir* p_dir, const struct mystr* p_base_dir_str,
   const struct mystr* p_option_str, const struct mystr* p_filter_str,
-  int is_verbose, int is_recurse);
+  int is_verbose);
 static int write_dir_list(struct mystr_list* p_dir_list, int remote_fd);
 
 void
@@ -62,8 +62,8 @@ vsf_ftpdataio_dispose_transfer_fd(struct vsf_session* p_sess)
   if (vsf_sysutil_retval_is_error(retval))
   {
     /* Do it again without blocking. */
-    vsf_sysutil_deactivate_linger(p_sess->data_fd);
-    vsf_sysutil_close(p_sess->data_fd);
+    vsf_sysutil_deactivate_linger_failok(p_sess->data_fd);
+    (void) vsf_sysutil_close_failok(p_sess->data_fd);
   }
   p_sess->data_fd = -1;
 }
@@ -72,8 +72,6 @@ int
 vsf_ftpdataio_get_pasv_fd(struct vsf_session* p_sess)
 {
   static struct vsf_sysutil_sockaddr* s_p_accept_addr = 0;
-  struct vsf_sysutil_ipv4addr cmd_conn_addr;
-  struct vsf_sysutil_ipv4addr remote_addr;
   int remote_fd = vsf_sysutil_accept_timeout(p_sess->pasv_listen_fd,
                                              &s_p_accept_addr,
                                              tunable_accept_timeout);
@@ -87,12 +85,10 @@ vsf_ftpdataio_get_pasv_fd(struct vsf_session* p_sess)
    * Reject the connection if it wasn't from the same IP as the
    * control connection.
    */
-  cmd_conn_addr = vsf_sysutil_sockaddr_get_ipaddr(p_sess->p_remote_addr);
-  remote_addr = vsf_sysutil_sockaddr_get_ipaddr(s_p_accept_addr);
   if (!tunable_pasv_promiscuous)
   {
-    if (vsf_sysutil_memcmp(cmd_conn_addr.data, remote_addr.data,
-                           sizeof(cmd_conn_addr)) != 0)
+    if (!vsf_sysutil_sockaddr_addr_equal(p_sess->p_remote_addr,
+                                         s_p_accept_addr))
     {
       vsf_cmdio_write(p_sess, FTP_BADSENDCONN, "Security: Bad IP connecting.");
       vsf_sysutil_close(remote_fd);
@@ -121,7 +117,18 @@ vsf_ftpdataio_get_port_fd(struct vsf_session* p_sess)
   }
   else
   {
-    remote_fd = vsf_sysutil_get_ipv4_sock();
+    remote_fd = vsf_sysutil_get_ipsock(p_sess->p_port_sockaddr);
+    if (vsf_sysutil_sockaddr_same_family(p_sess->p_port_sockaddr,
+                                         p_sess->p_local_addr))
+    {
+      static struct vsf_sysutil_sockaddr* s_p_addr;
+      vsf_sysutil_sockaddr_clone(&s_p_addr, p_sess->p_local_addr);
+      retval = vsf_sysutil_bind(remote_fd, s_p_addr);
+      if (retval != 0)
+      {
+        die("vsf_sysutil_bind");
+      }
+    }
   }
   retval = vsf_sysutil_connect_timeout(remote_fd, p_sess->p_port_sockaddr,
                                        tunable_connect_timeout);
@@ -239,7 +246,7 @@ vsf_ftpdataio_transfer_dir(struct vsf_session* p_sess, const int remote_fd,
                            int is_verbose)
 {
   return transfer_dir_internal(p_sess, remote_fd, p_dir, p_base_dir_str,
-                               p_option_str, p_filter_str, is_verbose, 0);
+                               p_option_str, p_filter_str, is_verbose);
 }
 
 static int
@@ -248,7 +255,7 @@ transfer_dir_internal(struct vsf_session* p_sess, const int remote_fd,
                       const struct mystr* p_base_dir_str,
                       const struct mystr* p_option_str,
                       const struct mystr* p_filter_str,
-                      int is_verbose, int is_recurse)
+                      int is_verbose)
 {
   struct mystr_list dir_list = INIT_STRLIST;
   struct mystr_list subdir_list = INIT_STRLIST;
@@ -311,8 +318,7 @@ transfer_dir_internal(struct vsf_session* p_sess, const int remote_fd,
         break;
       }
       retval = transfer_dir_internal(p_sess, remote_fd, p_subdir, &sub_str,
-                                     p_option_str, p_filter_str,
-                                     is_verbose, 1);
+                                     p_option_str, p_filter_str, is_verbose);
       vsf_sysutil_closedir(p_subdir);
       if (retval != 0)
       {
@@ -327,19 +333,10 @@ transfer_dir_internal(struct vsf_session* p_sess, const int remote_fd,
   str_free(&dir_prefix_str);
   if (!failed)
   {
-    if (!is_recurse)
-    {
-      vsf_cmdio_write(p_sess, FTP_TRANSFEROK, "Directory send OK.");
-    }
     return 0;
   }
   else
   {
-    if (!is_recurse)
-    {
-      vsf_cmdio_write(p_sess, FTP_BADSENDNET,
-                      "Failure writing network stream.");
-    }
     return -1;
   }
 }
@@ -406,6 +403,12 @@ do_file_send_ascii(struct vsf_session* p_sess, int net_fd, int file_fd)
   static char* p_readbuf;
   static char* p_asciibuf;
   struct vsf_transfer_ret ret_struct = { 0, 0 };
+  unsigned int chunk_size = VSFTP_DATA_BUFSIZE;
+  if (tunable_trans_chunk_size < VSFTP_DATA_BUFSIZE &&
+      tunable_trans_chunk_size >= 4096)
+  {
+    chunk_size = tunable_trans_chunk_size;
+  }
   if (p_readbuf == 0)
   {
     /* NOTE!! * 2 factor because we can double the data by doing our ASCII
@@ -417,7 +420,7 @@ do_file_send_ascii(struct vsf_session* p_sess, int net_fd, int file_fd)
   while (1)
   {
     unsigned int num_to_write;
-    int retval = vsf_sysutil_read(file_fd, p_readbuf, VSFTP_DATA_BUFSIZE);
+    int retval = vsf_sysutil_read(file_fd, p_readbuf, chunk_size);
     if (vsf_sysutil_retval_is_error(retval))
     {
       vsf_cmdio_write(p_sess, FTP_BADSENDFILE, "Failure reading local file.");
@@ -450,10 +453,10 @@ do_file_send_binary(struct vsf_session* p_sess, int net_fd, int file_fd)
 {
   static struct vsf_sysutil_statbuf* s_p_statbuf;
   int retval;
-  unsigned long bytes_to_send;
-  unsigned long init_file_offset;
-  unsigned long curr_file_offset;
-  unsigned long bytes_sent;
+  filesize_t bytes_to_send;
+  filesize_t init_file_offset;
+  filesize_t curr_file_offset;
+  filesize_t bytes_sent;
   unsigned int chunk_size;
   struct vsf_transfer_ret ret_struct = { 0, 0 };
   /* Work out how many bytes to send based on file size minus current offset */
@@ -461,6 +464,10 @@ do_file_send_binary(struct vsf_session* p_sess, int net_fd, int file_fd)
   vsf_sysutil_fstat(file_fd, &s_p_statbuf);
   bytes_to_send = vsf_sysutil_statbuf_get_size(s_p_statbuf);
   init_file_offset = vsf_sysutil_get_file_offset(file_fd);
+  if (init_file_offset < 0 || bytes_to_send < 0)
+  {
+    die("do_file_send_binary: negative file offset or send count");
+  }
   curr_file_offset = init_file_offset;
   /* Don't underflow if some bonehead sets a REST greater than the file size */
   if (init_file_offset > bytes_to_send)
@@ -474,6 +481,11 @@ do_file_send_binary(struct vsf_session* p_sess, int net_fd, int file_fd)
   if (p_sess->bw_rate_max)
   {
     chunk_size = VSFTP_DATA_BUFSIZE;
+    if (tunable_trans_chunk_size < VSFTP_DATA_BUFSIZE &&
+        tunable_trans_chunk_size >= 4096)
+    {
+      chunk_size = tunable_trans_chunk_size;
+    }
   }
   else
   {
@@ -507,13 +519,19 @@ do_file_recv(struct vsf_session* p_sess, int net_fd, int file_fd, int is_ascii)
   static char* p_recvbuf;
   unsigned int num_to_write;
   struct vsf_transfer_ret ret_struct = { 0, 0 };
+  unsigned int chunk_size = VSFTP_DATA_BUFSIZE;
   if (p_recvbuf == 0)
   {
     vsf_secbuf_alloc(&p_recvbuf, VSFTP_DATA_BUFSIZE);
   }
+  if (tunable_trans_chunk_size < VSFTP_DATA_BUFSIZE &&
+      tunable_trans_chunk_size >= 4096)
+  {
+    chunk_size = tunable_trans_chunk_size;
+  }
   while (1)
   {
-    int retval = vsf_sysutil_read(net_fd, p_recvbuf, VSFTP_DATA_BUFSIZE);
+    int retval = vsf_sysutil_read(net_fd, p_recvbuf, chunk_size);
     if (vsf_sysutil_retval_is_error(retval))
     {
       vsf_cmdio_write(p_sess, FTP_BADSENDNET,
