@@ -17,15 +17,17 @@
 #include "defs.h"
 #include "str.h"
 #include "sysstr.h"
-#include "dirchange.h"
+#include "banner.h"
 #include "sysutil.h"
 #include "logging.h"
 #include "sysdeputil.h"
+#include "ipv6parse.h"
+#include "access.h"
 
 /* Private local functions */
 static void handle_pwd(struct vsf_session* p_sess);
 static void handle_cwd(struct vsf_session* p_sess);
-static void handle_pasv(struct vsf_session* p_sess);
+static void handle_pasv(struct vsf_session* p_sess, int is_epsv);
 static void handle_retr(struct vsf_session* p_sess);
 static void handle_cdup(struct vsf_session* p_sess);
 static void handle_list(struct vsf_session* p_sess);
@@ -47,17 +49,26 @@ static void handle_site_chmod(struct vsf_session* p_sess,
                               struct mystr* p_arg_str);
 static void handle_site_umask(struct vsf_session* p_sess,
                               struct mystr* p_arg_str);
+static void handle_eprt(struct vsf_session* p_sess);
+static void handle_help(struct vsf_session* p_sess);
+static void handle_stou(struct vsf_session* p_sess);
+static void handle_stat(struct vsf_session* p_sess);
+static void handle_stat_file(struct vsf_session* p_sess);
 
 static int pasv_active(struct vsf_session* p_sess);
 static int port_active(struct vsf_session* p_sess);
 static void pasv_cleanup(struct vsf_session* p_sess);
 static void port_cleanup(struct vsf_session* p_sess);
-static void handle_dir_common(struct vsf_session* p_sess, int full_details);
+static void handle_dir_common(struct vsf_session* p_sess, int full_details,
+                              int stat_cmd);
 static void prepend_path_to_filename(struct mystr* p_str);
 static int get_remote_transfer_fd(struct vsf_session* p_sess);
 static int dispose_remote_transfer_fd(struct vsf_session* p_sess);
 static void handle_sigurg(void* p_private);
-static void handle_upload_common(struct vsf_session* p_sess, int is_append);
+static void handle_upload_common(struct vsf_session* p_sess, int is_append,
+                                 int is_unique);
+static void get_unique_filename(struct mystr* p_outstr,
+                                const struct mystr* p_base);
 
 void
 process_post_login(struct vsf_session* p_sess)
@@ -66,31 +77,23 @@ process_post_login(struct vsf_session* p_sess)
   {
     vsf_sysutil_set_umask(tunable_anon_umask);
     p_sess->bw_rate_max = tunable_anon_max_rate;
-    if (tunable_anon_root != 0)
-    {
-      (void) vsf_sysutil_chdir(tunable_anon_root);
-    }
   }
   else
   {
     vsf_sysutil_set_umask(tunable_local_umask);
     p_sess->bw_rate_max = tunable_local_max_rate;
-    if (tunable_local_root != 0)
-    {
-      (void) vsf_sysutil_chdir(tunable_local_root);
-    }
   }
   if (tunable_async_abor_enable)
   {
     vsf_sysutil_install_sighandler(kVSFSysUtilSigURG, handle_sigurg, p_sess);
-    vsf_sysutil_activate_oobinline(VSFTP_COMMAND_FD);
     vsf_sysutil_activate_sigurg(VSFTP_COMMAND_FD);
   }
   /* Handle any login message */
-  dir_changed(p_sess, FTP_LOGINOK);
-  vsf_cmdio_write(p_sess, FTP_LOGINOK, "Login successful. Have fun.");
+  vsf_banner_dir_changed(p_sess, FTP_LOGINOK);
+  vsf_cmdio_write(p_sess, FTP_LOGINOK, "Login successful.");
   while(1)
   {
+    int cmd_ok = 1;
     if (tunable_setproctitle_enable)
     {
       vsf_sysutil_setproctitle("IDLE");
@@ -112,7 +115,32 @@ process_post_login(struct vsf_session* p_sess)
       vsf_sysutil_setproctitle_str(&proctitle_str);
       str_free(&proctitle_str);
     }
-    if (str_equal_text(&p_sess->ftp_cmd_str, "QUIT"))
+    /* Test command against the allowed list.. */
+    if (tunable_cmds_allowed)
+    {
+      static struct mystr s_src_str;
+      static struct mystr s_rhs_str;
+      str_alloc_text(&s_src_str, tunable_cmds_allowed);
+      while (1)
+      {
+        str_split_char(&s_src_str, &s_rhs_str, ',');
+        if (str_isempty(&s_src_str))
+        {
+          cmd_ok = 0;
+          break;
+        }
+        else if (str_equal(&s_src_str, &p_sess->ftp_cmd_str))
+        {
+          break;
+        }
+        str_copy(&s_src_str, &s_rhs_str);
+      }
+    }
+    if (!cmd_ok)
+    {
+      vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "QUIT"))
     {
       vsf_cmdio_write(p_sess, FTP_GOODBYE, "Goodbye.");
       vsf_sysutil_exit(0);
@@ -133,17 +161,25 @@ process_post_login(struct vsf_session* p_sess)
       handle_cdup(p_sess);
     }
     else if (tunable_pasv_enable &&
-             str_equal_text(&p_sess->ftp_cmd_str, "PASV"))
+             !p_sess->epsv_all &&
+             (str_equal_text(&p_sess->ftp_cmd_str, "PASV") ||
+              str_equal_text(&p_sess->ftp_cmd_str, "P@SW")))
     {
-      handle_pasv(p_sess);
+      handle_pasv(p_sess, 0);
     }
-    else if (str_equal_text(&p_sess->ftp_cmd_str, "RETR"))
+    else if (tunable_pasv_enable &&
+             str_equal_text(&p_sess->ftp_cmd_str, "EPSV"))
+    {
+      handle_pasv(p_sess, 1);
+    }
+    else if (tunable_download_enable &&
+             str_equal_text(&p_sess->ftp_cmd_str, "RETR"))
     {
       handle_retr(p_sess);
     }
     else if (str_equal_text(&p_sess->ftp_cmd_str, "NOOP"))
     {
-      vsf_cmdio_write(p_sess, FTP_NOOPOK, "Mary had a little lamb.");
+      vsf_cmdio_write(p_sess, FTP_NOOPOK, "NOOP ok.");
     }
     else if (str_equal_text(&p_sess->ftp_cmd_str, "SYST"))
     {
@@ -151,9 +187,10 @@ process_post_login(struct vsf_session* p_sess)
     }
     else if (str_equal_text(&p_sess->ftp_cmd_str, "HELP"))
     {
-      vsf_cmdio_write(p_sess, FTP_BADHELP, "Sorry, I don't do help.");
+      handle_help(p_sess);
     }
-    else if (str_equal_text(&p_sess->ftp_cmd_str, "LIST"))
+    else if (tunable_dirlist_enable &&
+             str_equal_text(&p_sess->ftp_cmd_str, "LIST"))
     {
       handle_list(p_sess);
     }
@@ -162,6 +199,7 @@ process_post_login(struct vsf_session* p_sess)
       handle_type(p_sess);
     }
     else if (tunable_port_enable &&
+             !p_sess->epsv_all &&
              str_equal_text(&p_sess->ftp_cmd_str, "PORT"))
     {
       handle_port(p_sess);
@@ -208,7 +246,8 @@ process_post_login(struct vsf_session* p_sess)
     {
       handle_rnto(p_sess);
     }
-    else if (str_equal_text(&p_sess->ftp_cmd_str, "NLST"))
+    else if (tunable_dirlist_enable &&
+             str_equal_text(&p_sess->ftp_cmd_str, "NLST"))
     {
       handle_nlst(p_sess);
     }
@@ -221,12 +260,16 @@ process_post_login(struct vsf_session* p_sess)
     {
       handle_site(p_sess);
     }
-    else if (str_equal_text(&p_sess->ftp_cmd_str, "ABOR"))
+    /* Note - the weird ABOR string is checking for an async ABOR arriving
+     * without a SIGURG condition.
+     */
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "ABOR") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "\377\364\377\362ABOR"))
     {
-      vsf_cmdio_write(p_sess, FTP_ABOR_NOCONN, "Duh. No transfer to ABOR.");
+      vsf_cmdio_write(p_sess, FTP_ABOR_NOCONN, "No transfer to ABOR.");
     }
-    /* SECURITY: for now, no APPE in anonymous mode */
-    else if (!p_sess->is_anonymous &&
+    else if (tunable_write_enable &&
+             (tunable_anon_other_write_enable || !p_sess->is_anonymous) &&
              str_equal_text(&p_sess->ftp_cmd_str, "APPE"))
     {
       handle_appe(p_sess);
@@ -235,6 +278,105 @@ process_post_login(struct vsf_session* p_sess)
     {
       handle_mdtm(p_sess);
     }
+    else if (tunable_port_enable &&
+             str_equal_text(&p_sess->ftp_cmd_str, "EPRT"))
+    {
+      handle_eprt(p_sess);
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "STRU"))
+    {
+      str_upper(&p_sess->ftp_arg_str);
+      if (str_equal_text(&p_sess->ftp_arg_str, "F"))
+      {
+        vsf_cmdio_write(p_sess, FTP_STRUOK, "Structure set to F.");
+      }
+      else
+      {
+        vsf_cmdio_write(p_sess, FTP_BADSTRU, "Bad STRU command.");
+      }
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "MODE"))
+    {
+      str_upper(&p_sess->ftp_arg_str);
+      if (str_equal_text(&p_sess->ftp_arg_str, "S"))
+      {
+        vsf_cmdio_write(p_sess, FTP_MODEOK, "Mode set to S.");
+      }
+      else
+      {
+        vsf_cmdio_write(p_sess, FTP_BADMODE, "Bad MODE command.");
+      }
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "STOU"))
+    {
+      handle_stou(p_sess);
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "ALLO"))
+    {
+      vsf_cmdio_write(p_sess, FTP_ALLOOK, "ALLO command ignored.");
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "REIN"))
+    {
+      vsf_cmdio_write(p_sess, FTP_COMMANDNOTIMPL, "REIN not implemented.");
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "ACCT"))
+    {
+      vsf_cmdio_write(p_sess, FTP_COMMANDNOTIMPL, "ACCT not implemented.");
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "SMNT"))
+    {
+      vsf_cmdio_write(p_sess, FTP_COMMANDNOTIMPL, "SMNT not implemented.");
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "FEAT"))
+    {
+      vsf_cmdio_write_hyphen(p_sess, FTP_FEAT, "Features:");
+      vsf_cmdio_write_raw(p_sess, " MDTM\r\n");
+      vsf_cmdio_write_raw(p_sess, " REST STREAM\r\n");
+      vsf_cmdio_write_raw(p_sess, " SIZE\r\n");
+      vsf_cmdio_write(p_sess, FTP_FEAT, "End");
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "OPTS"))
+    {
+      vsf_cmdio_write(p_sess, FTP_BADOPTS, "Option not understood.");
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "STAT") &&
+             str_isempty(&p_sess->ftp_arg_str))
+    {
+      handle_stat(p_sess);
+    }
+    else if (tunable_dirlist_enable && 
+             str_equal_text(&p_sess->ftp_cmd_str, "STAT"))
+    {
+      handle_stat_file(p_sess);
+    }
+    else if (str_equal_text(&p_sess->ftp_cmd_str, "PASV") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "PORT") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "STOR") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "MKD") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "XMKD") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "RMD") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "XRMD") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "DELE") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "RNFR") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "RNTO") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "SITE") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "APPE") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "EPSV") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "EPRT") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "RETR") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "LIST") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "NLST") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "STOU") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "ALLO") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "REIN") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "ACCT") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "SMNT") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "FEAT") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "OPTS") ||
+             str_equal_text(&p_sess->ftp_cmd_str, "STAT"))
+    {
+      vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    }
     else
     {
       vsf_cmdio_write(p_sess, FTP_BADCMD, "Unknown command.");
@@ -260,11 +402,17 @@ handle_pwd(struct vsf_session* p_sess)
 static void
 handle_cwd(struct vsf_session* p_sess)
 {
-  int retval = str_chdir(&p_sess->ftp_arg_str);
+  int retval;
+  if (!vsf_access_check_file(&p_sess->ftp_arg_str))
+  {
+    vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    return;
+  }
+  retval = str_chdir(&p_sess->ftp_arg_str);
   if (retval == 0)
   {
     /* Handle any messages */
-    dir_changed(p_sess, FTP_CWDOK);
+    vsf_banner_dir_changed(p_sess, FTP_CWDOK);
     vsf_cmdio_write(p_sess, FTP_CWDOK, "Directory successfully changed.");
   }
   else
@@ -334,29 +482,53 @@ pasv_cleanup(struct vsf_session* p_sess)
 }
 
 static void
-handle_pasv(struct vsf_session* p_sess)
+handle_pasv(struct vsf_session* p_sess, int is_epsv)
 {
   static struct mystr s_pasv_res_str;
   static struct vsf_sysutil_sockaddr* s_p_sockaddr;
-  struct vsf_sysutil_ipv4port listen_port;
-  struct vsf_sysutil_ipv4addr listen_ipaddr;
   int bind_retries = 10;
+  unsigned short the_port = 0;
+  int is_ipv6 = vsf_sysutil_sockaddr_is_ipv6(p_sess->p_local_addr);
+  if (is_epsv && !str_isempty(&p_sess->ftp_arg_str))
+  {
+    int argval;
+    str_upper(&p_sess->ftp_arg_str);
+    if (str_equal_text(&p_sess->ftp_arg_str, "ALL"))
+    {
+      p_sess->epsv_all = 1;
+      vsf_cmdio_write(p_sess, FTP_EPSVALLOK, "EPSV ALL ok.");
+      return;
+    }
+    argval = vsf_sysutil_atoi(str_getbuf(&p_sess->ftp_arg_str));
+    if (!is_ipv6 || argval != 2)
+    {
+      vsf_cmdio_write(p_sess, FTP_EPSVBAD, "Bad network protocol.");
+      return;
+    }
+  }
   pasv_cleanup(p_sess);
   port_cleanup(p_sess);
-  p_sess->pasv_listen_fd = vsf_sysutil_get_ipv4_sock();
+  if (is_epsv && is_ipv6)
+  {
+    p_sess->pasv_listen_fd = vsf_sysutil_get_ipv6_sock();
+  }
+  else
+  {
+    p_sess->pasv_listen_fd = vsf_sysutil_get_ipv4_sock();
+  }
+  vsf_sysutil_activate_reuseaddr(p_sess->pasv_listen_fd);
   while (--bind_retries)
   {
     int retval;
-    unsigned short the_port;
     double scaled_port;
     /* IPPORT_RESERVED */
     unsigned short min_port = 1024;
     unsigned short max_port = 65535;
-    if (tunable_pasv_min_port > min_port && tunable_pasv_min_port < max_port)
+    if (tunable_pasv_min_port > min_port && tunable_pasv_min_port <= max_port)
     {
       min_port = tunable_pasv_min_port;
     }
-    if (tunable_pasv_max_port > min_port && tunable_pasv_max_port < max_port)
+    if (tunable_pasv_max_port >= min_port && tunable_pasv_max_port < max_port)
     {
       max_port = tunable_pasv_max_port;
     }
@@ -364,15 +536,11 @@ handle_pasv(struct vsf_session* p_sess)
     the_port <<= 8;
     the_port |= vsf_sysutil_get_random_byte();
     scaled_port = (double) min_port;
-    scaled_port += ((double) the_port / (double) 65535) *
-                   ((double) max_port - min_port);
+    scaled_port += ((double) the_port / (double) 65536) *
+                   ((double) max_port - min_port + 1);
     the_port = (unsigned short) scaled_port;
-    vsf_sysutil_sockaddr_alloc_ipv4(&s_p_sockaddr);
-    vsf_sysutil_sockaddr_set_port(s_p_sockaddr,
-                                  vsf_sysutil_ipv4port_from_int(the_port));
-    /* Bind to same address we got the incoming connect on */
-    vsf_sysutil_sockaddr_set_ipaddr(s_p_sockaddr,
-      vsf_sysutil_sockaddr_get_ipaddr(p_sess->p_local_addr));
+    vsf_sysutil_sockaddr_clone(&s_p_sockaddr, p_sess->p_local_addr);
+    vsf_sysutil_sockaddr_set_port(s_p_sockaddr, the_port);
     retval = vsf_sysutil_bind(p_sess->pasv_listen_fd, s_p_sockaddr);
     if (!vsf_sysutil_retval_is_error(retval))
     {
@@ -389,22 +557,40 @@ handle_pasv(struct vsf_session* p_sess)
     die("vsf_sysutil_bind");
   }
   vsf_sysutil_listen(p_sess->pasv_listen_fd, 1);
-  /* Get the address of the bound socket, for the port */
-  vsf_sysutil_getsockname(p_sess->pasv_listen_fd, &s_p_sockaddr);
-  listen_ipaddr = vsf_sysutil_sockaddr_get_ipaddr(s_p_sockaddr);
-  listen_port = vsf_sysutil_sockaddr_get_port(s_p_sockaddr);
+  if (is_epsv)
+  {
+    str_alloc_text(&s_pasv_res_str, "Entering Extended Passive Mode (|||");
+    str_append_ulong(&s_pasv_res_str, (unsigned long) the_port);
+    str_append_text(&s_pasv_res_str, "|)");
+    vsf_cmdio_write_str(p_sess, FTP_EPSVOK, &s_pasv_res_str);
+    return;
+  }
+  if (tunable_pasv_address != 0)
+  {
+    /* Report passive address as specified in configuration */
+    if (vsf_sysutil_inet_aton(tunable_pasv_address, s_p_sockaddr) == 0)
+    {
+      die("invalid pasv_address");
+    }
+  }
   str_alloc_text(&s_pasv_res_str, "Entering Passive Mode (");
-  str_append_ulong(&s_pasv_res_str, listen_ipaddr.data[0]);
+  if (!is_ipv6)
+  {
+    str_append_text(&s_pasv_res_str, vsf_sysutil_inet_ntop(s_p_sockaddr));
+  }
+  else
+  {
+    const void* p_v4addr = vsf_sysutil_sockaddr_ipv6_v4(s_p_sockaddr);
+    if (p_v4addr)
+    {
+      str_append_text(&s_pasv_res_str, vsf_sysutil_inet_ntoa(p_v4addr));
+    }
+  }
+  str_replace_char(&s_pasv_res_str, '.', ',');
   str_append_text(&s_pasv_res_str, ",");
-  str_append_ulong(&s_pasv_res_str, listen_ipaddr.data[1]);
+  str_append_ulong(&s_pasv_res_str, the_port >> 8);
   str_append_text(&s_pasv_res_str, ",");
-  str_append_ulong(&s_pasv_res_str, listen_ipaddr.data[2]);
-  str_append_text(&s_pasv_res_str, ",");
-  str_append_ulong(&s_pasv_res_str, listen_ipaddr.data[3]);
-  str_append_text(&s_pasv_res_str, ",");
-  str_append_ulong(&s_pasv_res_str, listen_port.data[0]);
-  str_append_text(&s_pasv_res_str, ",");
-  str_append_ulong(&s_pasv_res_str, listen_port.data[1]);
+  str_append_ulong(&s_pasv_res_str, the_port & 255);
   str_append_text(&s_pasv_res_str, ")");
   vsf_cmdio_write_str(p_sess, FTP_PASVOK, &s_pasv_res_str);
 }
@@ -419,7 +605,7 @@ handle_retr(struct vsf_session* p_sess)
   int remote_fd;
   int opened_file;
   int is_ascii = 0;
-  unsigned long offset = p_sess->restart_pos;
+  filesize_t offset = p_sess->restart_pos;
   p_sess->restart_pos = 0;
   if (!pasv_active(p_sess) && !port_active(p_sess))
   {
@@ -432,6 +618,11 @@ handle_retr(struct vsf_session* p_sess)
                     "No support for resume of ASCII transfer.");
     return;
   }
+  if (!vsf_access_check_file(&p_sess->ftp_arg_str))
+  {
+    vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    return;
+  }
   opened_file = str_open(&p_sess->ftp_arg_str, kVSFSysStrOpenReadOnly);
   if (vsf_sysutil_retval_is_error(opened_file))
   {
@@ -456,7 +647,7 @@ handle_retr(struct vsf_session* p_sess)
   /* Set the download offset (from REST) if any */
   if (offset != 0)
   {
-    (void) vsf_sysutil_lseek_to(opened_file, offset);
+    vsf_sysutil_lseek_to(opened_file, offset);
   }
   remote_fd = get_remote_transfer_fd(p_sess);
   if (vsf_sysutil_retval_is_error(remote_fd))
@@ -479,7 +670,8 @@ handle_retr(struct vsf_session* p_sess)
   str_append_text(&s_mark_str, " mode data connection for ");
   str_append_str(&s_mark_str, &p_sess->ftp_arg_str);
   str_append_text(&s_mark_str, " (");
-  str_append_ulong(&s_mark_str, vsf_sysutil_statbuf_get_size(s_p_statbuf));
+  str_append_filesize_t(&s_mark_str,
+                        vsf_sysutil_statbuf_get_size(s_p_statbuf));
   str_append_text(&s_mark_str, " bytes).");
   vsf_cmdio_write_str(p_sess, FTP_DATACONN, &s_mark_str);
   trans_ret = vsf_ftpdataio_transfer_file(p_sess, remote_fd,
@@ -505,11 +697,11 @@ file_close_out:
 static void
 handle_list(struct vsf_session* p_sess)
 {
-  handle_dir_common(p_sess, 1);
+  handle_dir_common(p_sess, 1, 0);
 }
 
 static void
-handle_dir_common(struct vsf_session* p_sess, int full_details)
+handle_dir_common(struct vsf_session* p_sess, int full_details, int stat_cmd)
 {
   static struct mystr s_option_str;
   static struct mystr s_filter_str;
@@ -518,11 +710,12 @@ handle_dir_common(struct vsf_session* p_sess, int full_details)
   int remote_fd;
   int dir_allow_read = 1;
   struct vsf_sysutil_dir* p_dir = 0;
+  int retval = 0;
   str_empty(&s_option_str);
   str_empty(&s_filter_str);
   /* By default open the current directory */
   str_alloc_text(&s_dir_name_str, ".");
-  if (!pasv_active(p_sess) && !port_active(p_sess))
+  if (!stat_cmd && !pasv_active(p_sess) && !port_active(p_sess))
   {
     vsf_cmdio_write(p_sess, FTP_BADSENDCONN, "Use PORT or PASV first.");
     return;
@@ -545,6 +738,11 @@ handle_dir_common(struct vsf_session* p_sess, int full_details)
   }
   if (!str_isempty(&s_filter_str))
   {
+    if (!vsf_access_check_file(&s_filter_str))
+    {
+      vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+      return;
+    }
     /* First check - is it an outright directory, as in "ls /pub" */
     p_dir = str_opendir(&s_filter_str);
     if (p_dir != 0)
@@ -580,12 +778,27 @@ handle_dir_common(struct vsf_session* p_sess, int full_details)
     p_dir = str_opendir(&s_dir_name_str);
   }
   /* Fine, do it */
-  remote_fd = get_remote_transfer_fd(p_sess);
+  if (stat_cmd)
+  {
+    remote_fd = VSFTP_COMMAND_FD;
+    str_append_char(&s_option_str, 'a');
+  }
+  else
+  {
+    remote_fd = get_remote_transfer_fd(p_sess);
+  }
   if (vsf_sysutil_retval_is_error(remote_fd))
   {
     goto dir_close_out;
   }
-  vsf_cmdio_write(p_sess, FTP_DATACONN, "Here comes the directory listing.");
+  if (stat_cmd)
+  {
+    vsf_cmdio_write_hyphen(p_sess, FTP_STATFILE_OK, "Status follows:");
+  }
+  else
+  {
+    vsf_cmdio_write(p_sess, FTP_DATACONN, "Here comes the directory listing.");
+  }
   if (p_sess->is_anonymous && p_dir && tunable_anon_world_readable_only)
   {
     vsf_sysutil_dir_stat(p_dir, &s_p_dirstat);
@@ -594,25 +807,43 @@ handle_dir_common(struct vsf_session* p_sess, int full_details)
       dir_allow_read = 0;
     }
   }
-  if (p_dir == 0 || !dir_allow_read)
+  if (p_dir != 0 && dir_allow_read)
+  {
+    retval = vsf_ftpdataio_transfer_dir(p_sess, remote_fd, p_dir,
+                                        &s_dir_name_str, &s_option_str,
+                                        &s_filter_str, full_details);
+  }
+  if (stat_cmd)
+  {
+    vsf_cmdio_write(p_sess, FTP_STATFILE_OK, "End of status");
+  }
+  else if (p_dir == 0 || !dir_allow_read)
   {
     vsf_cmdio_write(p_sess, FTP_TRANSFEROK,
                     "Transfer done (but failed to open directory).");
   }
+  else if (retval == 0)
+  {
+    vsf_cmdio_write(p_sess, FTP_TRANSFEROK, "Directory send OK.");
+  }
   else
   {
-    (void) vsf_ftpdataio_transfer_dir(p_sess, remote_fd, p_dir,
-                                      &s_dir_name_str, &s_option_str,
-                                      &s_filter_str, full_details);
+    vsf_cmdio_write(p_sess, FTP_BADSENDNET, "Failure writing network stream.");
+  }
+  if (!stat_cmd)
+  {
+    (void) dispose_remote_transfer_fd(p_sess);
   }
-  (void) dispose_remote_transfer_fd(p_sess);
 dir_close_out:
   if (p_dir)
   {
     vsf_sysutil_closedir(p_dir);
   }
-  port_cleanup(p_sess);
-  pasv_cleanup(p_sess);
+  if (!stat_cmd)
+  {
+    port_cleanup(p_sess);
+    pasv_cleanup(p_sess);
+  }
 }
 
 static void
@@ -624,13 +855,13 @@ handle_type(struct vsf_session* p_sess)
       str_equal_text(&p_sess->ftp_arg_str, "L 8"))
   {
     p_sess->is_ascii = 0;
-    vsf_cmdio_write(p_sess, FTP_TYPEOK, "Binary it is, then.");
+    vsf_cmdio_write(p_sess, FTP_TYPEOK, "Switching to Binary mode.");
   }
   else if (str_equal_text(&p_sess->ftp_arg_str, "A") ||
            str_equal_text(&p_sess->ftp_arg_str, "A N"))
   {
     p_sess->is_ascii = 1;
-    vsf_cmdio_write(p_sess, FTP_TYPEOK, "ASCII tastes bad, dude.");
+    vsf_cmdio_write(p_sess, FTP_TYPEOK, "Switching to ASCII mode.");
   }
   else
   {
@@ -642,12 +873,9 @@ static void
 handle_port(struct vsf_session* p_sess)
 {
   static struct mystr s_tmp_str;
-  struct vsf_sysutil_ipv4addr the_addr;
-  struct vsf_sysutil_ipv4port the_port;
+  unsigned short the_port;
   unsigned char vals[6];
   int i;
-  struct vsf_sysutil_ipv4addr remote_addr =
-    vsf_sysutil_sockaddr_get_ipaddr(p_sess->p_remote_addr);
   pasv_cleanup(p_sess);
   port_cleanup(p_sess);
   str_copy(&s_tmp_str, &p_sess->ftp_arg_str);
@@ -677,23 +905,26 @@ handle_port(struct vsf_session* p_sess)
      */
     str_copy(&s_tmp_str, &s_rhs_comma_str);
   }
-  vsf_sysutil_memcpy(the_addr.data, vals, sizeof(the_addr.data));
-  vsf_sysutil_memcpy(the_port.data, &vals[4], sizeof(the_port.data));
+  the_port = vals[4] << 8;
+  the_port |= vals[5];
+  vsf_sysutil_sockaddr_alloc_ipv4(&p_sess->p_port_sockaddr);
+  vsf_sysutil_sockaddr_set_ipv4addr(p_sess->p_port_sockaddr, vals);
+  vsf_sysutil_sockaddr_set_port(p_sess->p_port_sockaddr, the_port);
   /* SECURITY:
    * 1) Reject requests not connecting to the control socket IP
    * 2) Reject connects to privileged ports
    */
-  if (vsf_sysutil_memcmp(the_addr.data, remote_addr.data,
-                         sizeof(the_addr.data)) != 0 ||
-      vsf_sysutil_is_port_reserved(the_port))
+  if (!tunable_port_promiscuous)
   {
-    vsf_cmdio_write(p_sess, FTP_BADCMD, "Illegal PORT command.");
-    port_cleanup(p_sess);
-    return;
+    if (!vsf_sysutil_sockaddr_addr_equal(p_sess->p_remote_addr,
+                                         p_sess->p_port_sockaddr) ||
+        vsf_sysutil_is_port_reserved(the_port))
+    {
+      vsf_cmdio_write(p_sess, FTP_BADCMD, "Illegal PORT command.");
+      port_cleanup(p_sess);
+      return;
+    }
   }
-  vsf_sysutil_sockaddr_alloc_ipv4(&p_sess->p_port_sockaddr);
-  vsf_sysutil_sockaddr_set_port(p_sess->p_port_sockaddr, the_port);
-  vsf_sysutil_sockaddr_set_ipaddr(p_sess->p_port_sockaddr, the_addr);
   vsf_cmdio_write(p_sess, FTP_PORTOK,
                   "PORT command successful. Consider using PASV.");
 }
@@ -701,41 +932,54 @@ handle_port(struct vsf_session* p_sess)
 static void
 handle_stor(struct vsf_session* p_sess)
 {
-  handle_upload_common(p_sess, 0);
+  handle_upload_common(p_sess, 0, 0);
 }
 
 static void
-handle_upload_common(struct vsf_session* p_sess, int is_append)
+handle_upload_common(struct vsf_session* p_sess, int is_append, int is_unique)
 {
+  static struct mystr s_filename;
+  struct mystr* p_filename;
   struct vsf_transfer_ret trans_ret;
   int new_file_fd;
   int remote_fd;
   int retval;
-  unsigned long offset = p_sess->restart_pos;
+  filesize_t offset = p_sess->restart_pos;
   p_sess->restart_pos = 0;
   if (!pasv_active(p_sess) && !port_active(p_sess))
   {
     vsf_cmdio_write(p_sess, FTP_BADSENDCONN, "Use PORT or PASV first.");
     return;
   }
+  p_filename = &p_sess->ftp_arg_str;
+  if (is_unique)
+  {
+    get_unique_filename(&s_filename, p_filename);
+    p_filename = &s_filename;
+  }
+  if (!vsf_access_check_file(p_filename))
+  {
+    vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    return;
+  }
   /* NOTE - actual file permissions will be governed by the tunable umask */
   /* XXX - do we care about race between create and chown() of anonymous
    * upload?
    */
-  if (p_sess->is_anonymous)
+  if (is_unique || (p_sess->is_anonymous && !tunable_anon_other_write_enable))
   {
-    new_file_fd = str_create(&p_sess->ftp_arg_str);
+    new_file_fd = str_create(p_filename);
   }
   else
   {
     /* For non-anonymous, allow open() to overwrite or append existing files */
     if (!is_append && offset == 0)
     {
-      new_file_fd = str_create_overwrite(&p_sess->ftp_arg_str);
+      new_file_fd = str_create_overwrite(p_filename);
     }
     else
     {
-      new_file_fd = str_create_append(&p_sess->ftp_arg_str);
+      new_file_fd = str_create_append(p_filename);
     }
   }
   if (vsf_sysutil_retval_is_error(new_file_fd))
@@ -746,6 +990,7 @@ handle_upload_common(struct vsf_session* p_sess, int is_append)
   /* Are we required to chown() this file for security? */
   if (p_sess->is_anonymous && tunable_chown_uploads)
   {
+    vsf_sysutil_fchmod(new_file_fd, 0600);
     if (tunable_one_process_model)
     {
       vsf_one_process_chown_upload(p_sess, new_file_fd);
@@ -758,15 +1003,25 @@ handle_upload_common(struct vsf_session* p_sess, int is_append)
   if (!is_append && offset != 0)
   {
     /* XXX - warning, allows seek past end of file! Check for seek > size? */
-    (void) vsf_sysutil_lseek_to(new_file_fd, offset);
+    vsf_sysutil_lseek_to(new_file_fd, offset);
   }
   remote_fd = get_remote_transfer_fd(p_sess);
   if (vsf_sysutil_retval_is_error(remote_fd))
   {
     goto port_pasv_cleanup_out;
   }
-  vsf_cmdio_write(p_sess, FTP_DATACONN,
-                  "Go ahead make my day^W^W^Wsend me the data.");
+  if (is_unique)
+  {
+    struct mystr resp_str = INIT_MYSTR;
+    str_alloc_text(&resp_str, "FILE: ");
+    str_append_str(&resp_str, p_filename);
+    vsf_cmdio_write_str(p_sess, FTP_DATACONN, &resp_str);
+    str_free(&resp_str);
+  }
+  else
+  {
+    vsf_cmdio_write(p_sess, FTP_DATACONN, "Ok to send data.");
+  }
   vsf_log_start_entry(p_sess, kVSFLogEntryUpload);
   str_copy(&p_sess->log_str, &p_sess->ftp_arg_str);
   prepend_path_to_filename(&p_sess->log_str);
@@ -802,6 +1057,11 @@ static void
 handle_mkd(struct vsf_session* p_sess)
 {
   int retval;
+  if (!vsf_access_check_file(&p_sess->ftp_arg_str))
+  {
+    vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    return;
+  }
   vsf_log_start_entry(p_sess, kVSFLogEntryMkdir);
   str_copy(&p_sess->log_str, &p_sess->ftp_arg_str);
   prepend_path_to_filename(&p_sess->log_str);
@@ -833,7 +1093,13 @@ handle_mkd(struct vsf_session* p_sess)
 static void
 handle_rmd(struct vsf_session* p_sess)
 {
-  int retval = str_rmdir(&p_sess->ftp_arg_str);
+  int retval;
+  if (!vsf_access_check_file(&p_sess->ftp_arg_str))
+  {
+    vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    return;
+  }
+  retval = str_rmdir(&p_sess->ftp_arg_str);
   if (retval != 0)
   {
     vsf_cmdio_write(p_sess, FTP_FILEFAIL,
@@ -849,7 +1115,13 @@ handle_rmd(struct vsf_session* p_sess)
 static void
 handle_dele(struct vsf_session* p_sess)
 {
-  int retval = str_unlink(&p_sess->ftp_arg_str);
+  int retval;
+  if (!vsf_access_check_file(&p_sess->ftp_arg_str))
+  {
+    vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    return;
+  }
+  retval = str_unlink(&p_sess->ftp_arg_str);
   if (retval != 0)
   {
     vsf_cmdio_write(p_sess, FTP_FILEFAIL, "Delete operation failed.");
@@ -863,13 +1135,17 @@ handle_dele(struct vsf_session* p_sess)
 static void
 handle_rest(struct vsf_session* p_sess)
 {
-  long val = str_atol(&p_sess->ftp_arg_str);
+  static struct mystr s_rest_str;
+  filesize_t val = str_a_to_filesize_t(&p_sess->ftp_arg_str);
   if (val < 0)
   {
     val = 0;
   }
-  p_sess->restart_pos = (unsigned long) val;
-  vsf_cmdio_write(p_sess, FTP_RESTOK, "Restart position accepted.");
+  p_sess->restart_pos = val;
+  str_alloc_text(&s_rest_str, "Restart position accepted (");
+  str_append_filesize_t(&s_rest_str, val);
+  str_append_text(&s_rest_str, ").");
+  vsf_cmdio_write_str(p_sess, FTP_RESTOK, &s_rest_str);
 }
 
 static void
@@ -879,6 +1155,11 @@ handle_rnfr(struct vsf_session* p_sess)
   int retval;
   /* Clear old value */
   str_free(&p_sess->rnfr_filename_str);
+  if (!vsf_access_check_file(&p_sess->ftp_arg_str))
+  {
+    vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    return;
+  }
   /* Does it exist? */
   retval = str_stat(&p_sess->ftp_arg_str, &p_statbuf);
   if (retval == 0)
@@ -901,7 +1182,12 @@ handle_rnto(struct vsf_session* p_sess)
   if (str_isempty(&p_sess->rnfr_filename_str))
   {
     vsf_cmdio_write(p_sess, FTP_NEEDRNFR,
-                    "Dude, get it sorted, I need RNFR first.");
+                    "RNFR required first.");
+    return;
+  }
+  if (!vsf_access_check_file(&p_sess->ftp_arg_str))
+  {
+    vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
     return;
   }
   /* NOTE - might overwrite destination file. Not a concern because the same
@@ -923,7 +1209,7 @@ handle_rnto(struct vsf_session* p_sess)
 static void
 handle_nlst(struct vsf_session* p_sess)
 {
-  handle_dir_common(p_sess, 0);
+  handle_dir_common(p_sess, 0, 0);
 }
 
 static void
@@ -1031,7 +1317,13 @@ handle_size(struct vsf_session* p_sess)
    * I will not do it because it is a potential I/O DoS.
    */
   static struct vsf_sysutil_statbuf* s_p_statbuf;
-  int retval = str_stat(&p_sess->ftp_arg_str, &s_p_statbuf);
+  int retval;
+  if (!vsf_access_check_file(&p_sess->ftp_arg_str))
+  {
+    vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    return;
+  }
+  retval = str_stat(&p_sess->ftp_arg_str, &s_p_statbuf);
   if (retval != 0 || !vsf_sysutil_statbuf_is_regfile(s_p_statbuf))
   {
     vsf_cmdio_write(p_sess, FTP_FILEFAIL, "Could not get file size.");
@@ -1039,8 +1331,8 @@ handle_size(struct vsf_session* p_sess)
   else
   {
     static struct mystr s_size_res_str;
-    str_alloc_ulong(&s_size_res_str,
-                    vsf_sysutil_statbuf_get_size(s_p_statbuf));
+    str_alloc_filesize_t(&s_size_res_str,
+                         vsf_sysutil_statbuf_get_size(s_p_statbuf));
     vsf_cmdio_write_str(p_sess, FTP_SIZEOK, &s_size_res_str);
   }
 }
@@ -1053,6 +1345,7 @@ handle_site(struct vsf_session* p_sess)
   str_split_char(&p_sess->ftp_arg_str, &s_site_args_str, ' ');
   str_upper(&p_sess->ftp_arg_str);
   if (tunable_write_enable &&
+      tunable_chmod_enable &&
       str_equal_text(&p_sess->ftp_arg_str, "CHMOD"))
   {
     handle_site_chmod(p_sess, &s_site_args_str);
@@ -1061,6 +1354,10 @@ handle_site(struct vsf_session* p_sess)
   {
     handle_site_umask(p_sess, &s_site_args_str);
   }
+  else if (str_equal_text(&p_sess->ftp_arg_str, "HELP"))
+  {
+    vsf_cmdio_write(p_sess, FTP_SITEHELP, "CHMOD UMASK HELP");
+  }
   else
   {
     vsf_cmdio_write(p_sess, FTP_BADCMD, "Unknown SITE command.");
@@ -1084,6 +1381,11 @@ handle_site_chmod(struct vsf_session* p_sess, struct mystr* p_arg_str)
     vsf_cmdio_write(p_sess, FTP_BADCMD, "SITE CHMOD needs 2 arguments.");
     return;
   }
+  if (!vsf_access_check_file(&s_chmod_file_str))
+  {
+    vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    return;
+  }
   /* Don't worry - our chmod() implementation only allows 0 - 0777 */
   perms = str_octal_to_uint(p_arg_str);
   retval = str_chmod(&s_chmod_file_str, perms);
@@ -1123,14 +1425,20 @@ handle_site_umask(struct vsf_session* p_sess, struct mystr* p_arg_str)
 static void
 handle_appe(struct vsf_session* p_sess)
 {
-  handle_upload_common(p_sess, 1);
+  handle_upload_common(p_sess, 1, 0);
 }
 
 static void
 handle_mdtm(struct vsf_session* p_sess)
 {
   static struct vsf_sysutil_statbuf* s_p_statbuf;
-  int retval = str_stat(&p_sess->ftp_arg_str, &s_p_statbuf);
+  int retval;
+  if (!vsf_access_check_file(&p_sess->ftp_arg_str))
+  {
+    vsf_cmdio_write(p_sess, FTP_NOPERM, "Permission denied.");
+    return;
+  }
+  retval = str_stat(&p_sess->ftp_arg_str, &s_p_statbuf);
   if (retval != 0 || !vsf_sysutil_statbuf_is_regfile(s_p_statbuf))
   {
     vsf_cmdio_write(p_sess, FTP_FILEFAIL,
@@ -1140,8 +1448,176 @@ handle_mdtm(struct vsf_session* p_sess)
   {
     static struct mystr s_mdtm_res_str;
     str_alloc_text(&s_mdtm_res_str,
-                   vsf_sysutil_statbuf_get_numeric_date(s_p_statbuf));
+                   vsf_sysutil_statbuf_get_numeric_date(
+                     s_p_statbuf, tunable_use_localtime));
     vsf_cmdio_write_str(p_sess, FTP_MDTMOK, &s_mdtm_res_str);
   }
 }
 
+static void
+handle_eprt(struct vsf_session* p_sess)
+{
+  static struct mystr s_part1_str;
+  static struct mystr s_part2_str;
+  int proto;
+  int port;
+  const unsigned char* p_raw_addr;
+  int is_ipv6 = vsf_sysutil_sockaddr_is_ipv6(p_sess->p_local_addr);
+  port_cleanup(p_sess);
+  pasv_cleanup(p_sess);
+  str_copy(&s_part1_str, &p_sess->ftp_arg_str);
+  str_split_char(&s_part1_str, &s_part2_str, '|');
+  if (!str_isempty(&s_part1_str))
+  {
+    goto bad_eprt;
+  }
+  /* Split out the protocol and check it */
+  str_split_char(&s_part2_str, &s_part1_str, '|');
+  proto = str_atoi(&s_part2_str);
+  if (!is_ipv6 || proto != 2)
+  {
+    vsf_cmdio_write(p_sess, FTP_BADCMD, "Bad EPRT protocol.");
+    return;
+  }
+  /* Split out address and parse it */
+  str_split_char(&s_part1_str, &s_part2_str, '|');
+  p_raw_addr = vsf_sysutil_parse_ipv6(&s_part1_str);
+  if (!p_raw_addr)
+  {
+    goto bad_eprt;
+  }
+  /* Split out port and parse it */
+  str_split_char(&s_part2_str, &s_part1_str, '|');
+  if (!str_isempty(&s_part1_str) || str_isempty(&s_part2_str))
+  {
+    goto bad_eprt;
+  }
+  port = str_atoi(&s_part2_str);
+  if (port < 0 || port > 65535)
+  {
+    goto bad_eprt;
+  }
+  vsf_sysutil_sockaddr_alloc_ipv6(&p_sess->p_port_sockaddr);
+  vsf_sysutil_sockaddr_set_ipv6addr(p_sess->p_port_sockaddr, p_raw_addr);
+  vsf_sysutil_sockaddr_set_port(p_sess->p_port_sockaddr, (unsigned short)port);
+  /* SECURITY:
+   * 1) Reject requests not connecting to the control socket IP
+   * 2) Reject connects to privileged ports
+   */
+  if (!tunable_port_promiscuous)
+  {
+    if (!vsf_sysutil_sockaddr_addr_equal(p_sess->p_remote_addr,
+                                         p_sess->p_port_sockaddr) ||
+        vsf_sysutil_is_port_reserved(port))
+    {
+      vsf_cmdio_write(p_sess, FTP_BADCMD, "Illegal EPRT command.");
+      port_cleanup(p_sess);
+      return;
+    }
+  }
+  vsf_cmdio_write(p_sess, FTP_EPRTOK,
+                  "EPRT command successful. Consider using EPSV.");
+  return;
+bad_eprt:
+  vsf_cmdio_write(p_sess, FTP_BADCMD, "Bad EPRT command.");
+}
+
+static void
+handle_help(struct vsf_session* p_sess)
+{
+  vsf_cmdio_write_hyphen(p_sess, FTP_HELP,
+                         "The following commands are recognized.");
+  vsf_cmdio_write_raw(p_sess,
+" ABOR ACCT ALLO APPE CDUP CWD  DELE EPRT EPSV FEAT HELP LIST MDTM MKD\r\n");
+  vsf_cmdio_write_raw(p_sess,
+" MODE NLST NOOP OPTS PASS PASV PORT PWD  QUIT REIN REST RETR RMD  RNFR\r\n");
+  vsf_cmdio_write_raw(p_sess,
+" RNTO SITE SIZE SMNT STAT STOR STOU STRU SYST TYPE USER XCUP XCWD XMKD\r\n");
+  vsf_cmdio_write_raw(p_sess,
+" XPWD XRMD\r\n");
+  vsf_cmdio_write(p_sess, FTP_HELP, "Help OK.");
+}
+
+static void
+handle_stou(struct vsf_session* p_sess)
+{
+  handle_upload_common(p_sess, 0, 1);
+}
+
+static void
+get_unique_filename(struct mystr* p_outstr, const struct mystr* p_base_str)
+{
+  /* Use silly wu-ftpd algorithm for compatibility */
+  static struct vsf_sysutil_statbuf* s_p_statbuf;
+  unsigned int suffix = 1;
+  int retval;
+  while (1)
+  {
+    str_copy(p_outstr, p_base_str);
+    str_append_char(p_outstr, '.');
+    str_append_ulong(p_outstr, suffix);
+    retval = str_stat(p_outstr, &s_p_statbuf);
+    if (vsf_sysutil_retval_is_error(retval))
+    {
+      return;
+    }
+    ++suffix;
+  }
+}
+
+static void
+handle_stat(struct vsf_session* p_sess)
+{
+  vsf_cmdio_write_hyphen(p_sess, FTP_STATOK, "FTP server status:");
+  vsf_cmdio_write_raw(p_sess, "     Connected to ");
+  vsf_cmdio_write_raw(p_sess, str_getbuf(&p_sess->remote_ip_str));
+  vsf_cmdio_write_raw(p_sess, "\r\n");
+  vsf_cmdio_write_raw(p_sess, "     Logged in as ");
+  vsf_cmdio_write_raw(p_sess, str_getbuf(&p_sess->user_str));
+  vsf_cmdio_write_raw(p_sess, "\r\n");
+  vsf_cmdio_write_raw(p_sess, "     TYPE: ");
+  if (p_sess->is_ascii)
+  {
+    vsf_cmdio_write_raw(p_sess, "ASCII\r\n");
+  }
+  else
+  {
+    vsf_cmdio_write_raw(p_sess, "BINARY\r\n");
+  }
+  if (p_sess->bw_rate_max == 0)
+  {
+    vsf_cmdio_write_raw(p_sess, "     No session bandwidth limit\r\n");
+  }
+  else
+  {
+    vsf_cmdio_write_raw(p_sess, "     Session bandwidth limit in byte/s is ");
+    vsf_cmdio_write_raw(p_sess, vsf_sysutil_ulong_to_str(p_sess->bw_rate_max));
+    vsf_cmdio_write_raw(p_sess, "\r\n");
+  }
+  if (tunable_idle_session_timeout == 0)
+  {
+    vsf_cmdio_write_raw(p_sess, "     No session timeout\r\n");
+  }
+  else
+  {
+    vsf_cmdio_write_raw(p_sess, "     Session timeout in seconds is ");
+    vsf_cmdio_write_raw(p_sess,
+      vsf_sysutil_ulong_to_str(tunable_idle_session_timeout));
+    vsf_cmdio_write_raw(p_sess, "\r\n");
+  }
+  if (p_sess->num_clients > 0)
+  {
+    vsf_cmdio_write_raw(p_sess, "     At session startup, client count was ");
+    vsf_cmdio_write_raw(p_sess, vsf_sysutil_ulong_to_str(p_sess->num_clients));
+    vsf_cmdio_write_raw(p_sess, "\r\n");
+  }
+  vsf_cmdio_write_raw(p_sess, "     vsFTPd - secure, fast, stable\r\n");
+  vsf_cmdio_write(p_sess, FTP_STATOK, "End of status");
+}
+
+static void
+handle_stat_file(struct vsf_session* p_sess)
+{
+  handle_dir_common(p_sess, 1, 1);
+}
+
