@@ -36,6 +36,26 @@ int tunable_log_ftp_protocol = 0;
 int tunable_guest_enable = 0;
 int tunable_userlist_enable = 0;
 int tunable_userlist_deny = 1;
+int tunable_use_localtime = 0;
+int tunable_check_shell = 1;
+int tunable_hide_ids = 0;
+int tunable_listen = 0;
+int tunable_port_promiscuous = 0;
+int tunable_passwd_chroot_enable = 0;
+int tunable_no_anon_password = 0;
+int tunable_tcp_wrappers = 0;
+int tunable_use_sendfile = 1;
+int tunable_force_dot_files = 0;
+int tunable_listen_ipv6 = 0;
+int tunable_dual_log_enable = 0;
+int tunable_syslog_enable = 0;
+int tunable_background = 0;
+int tunable_virtual_use_local_privs = 0;
+int tunable_session_support = 0;
+int tunable_download_enable = 1;
+int tunable_dirlist_enable = 1;
+int tunable_chmod_enable = 1;
+int tunable_secure_email_list_enable = 0;
 
 unsigned int tunable_accept_timeout = 60;
 unsigned int tunable_connect_timeout = 60;
@@ -49,13 +69,20 @@ unsigned int tunable_pasv_min_port = 5001;
 unsigned int tunable_pasv_max_port = 0;
 unsigned int tunable_anon_max_rate = 0;
 unsigned int tunable_local_max_rate = 0;
+/* IPPORT_FTP */
+unsigned int tunable_listen_port = 21;
+unsigned int tunable_max_clients = 0;
+/* -rw-rw-rw- */
+unsigned int tunable_file_open_mode = 0666;
+unsigned int tunable_max_per_ip = 0;
+unsigned int tunable_trans_chunk_size = 0;
 
 const char* tunable_secure_chroot_dir = "/usr/share/empty";
 const char* tunable_ftp_username = "ftp";
 const char* tunable_chown_username = "root";
-const char* tunable_xferlog_file = "/var/log/vsftpd.log";
+const char* tunable_xferlog_file = "/var/log/xferlog";
+const char* tunable_vsftpd_log_file = "/var/log/vsftpd.log";
 const char* tunable_message_file = ".message";
-/* XXX -> "secure"? */
 const char* tunable_nopriv_user = "nobody";
 const char* tunable_ftpd_banner = 0;
 const char* tunable_banned_email_file = "/etc/vsftpd.banned_emails";
@@ -65,4 +92,14 @@ const char* tunable_guest_username = "ftp";
 const char* tunable_userlist_file = "/etc/vsftpd.user_list";
 const char* tunable_anon_root = 0;
 const char* tunable_local_root = 0;
+const char* tunable_banner_file = 0;
+const char* tunable_pasv_address = 0;
+const char* tunable_listen_address = 0;
+const char* tunable_user_config_dir = 0;
+const char* tunable_listen_address6 = 0;
+const char* tunable_cmds_allowed = 0;
+const char* tunable_hide_file = 0;
+const char* tunable_deny_file = 0;
+const char* tunable_user_sub_token = 0;
+const char* tunable_email_password_file = "/etc/vsftpd.email_passwords";
 
